//PROFILE-NO
unit Misc;

// *****************************************************************************
// * Copyright 2003-2004 mxbee                                                 *
// *****************************************************************************
// * This program is free software; you can redistribute it and/or modify      *
// * it under the terms of the GNU General Public License as published by      *
// * the Free Software Foundation; either version 2 of the License, or         *
// * (at your option) any later version.                                       *
// *                                                                           *
// * This program is distributed in the hope that it will be useful,           *
// * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
// * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
// * GNU General Public License for more details.                              *
// *                                                                           *
// * You should have received a copy of the GNU General Public License         *
// * along with this program; if not, write to the Free Software               *
// * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
// *****************************************************************************

{$INCLUDE CompilerOpts.pas}

interface

uses Windows,SysUtils,Forms,Dialogs,IniFiles,Classes;

{-DEFINE DEBUG_SSMIF_DISABLE_QUICKREAD}
{-DEFINE DEBUG_SSMIF_DISABLE_QUICKWRITE}

type
  TSafeSaveMemIniFile = class (TMemIniFile)
    FQuickWrite:        Boolean;
    FQuickWriteSection: String;
    FQuickWriteStrings: TStrings;
    FQuickRead:         Boolean;
    FQuickReadSection:  String;
    FQuickReadStrings:  TStrings;
    FQuickReadIndex:    Integer;
  public
    destructor Destroy; override;
    procedure UpdateFile; override;
    procedure WriteString(const Section, Ident, Value: String); override;
    function  ReadString(const Section, Ident, Default: String): String; override;
    procedure QuickWriteBegin(const NewSection: String = '');
    procedure QuickWriteNewSection(const NewSection: String);
    procedure QuickWriteEnd;
    procedure QuickReadBegin(const NewSection: String = '');
    procedure QuickReadNewSection(const NewSection: String);
    procedure QuickReadEnd;
  end;

  TCachedFile = class;
  ECachedFile = class (Exception);

  TCachedFileList = class
  private
    FList: TList;
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Add(ACachedFile: TCachedFile);
    procedure   Remove(ACachedFile: TCachedFile);
    function    OpenFile(AFilename: String; MustExist: Boolean): TCachedFile;
    function    FlushAll(CloseOldFiles: Boolean = False): Integer;
    procedure   CloseFile(AFilename: String; FlushIt: Boolean);
    procedure   CloseAll(FlushIt: Boolean);
  end;

  TCachedFile = class
  private
    FFilename:  String;
    FMemStream: TMemoryStream;
    FDirty:     Boolean;
    FCacheList: TCachedFileList;
    FLastAccess: TDateTime;
    function    GetSize: Int64;
  public
    constructor Create(ACacheList: TCachedFileList; AFilename: String; InitFromFile: Boolean);
    destructor  Destroy; override;
    procedure   SetContents(PData: Pointer; Count: Longint);
    procedure   GetContents(PData: Pointer; Count: Longint);
    function    FlushToFile: Boolean;
    property    Filename: String read FFilename;
    property    Size: Int64 read GetSize;
    property    LastAccess: TDateTime read FLastAccess;
  end;

var
  gCachedFiles: TCachedFileList = nil;
    
procedure AdjustFormSize(Form: TForm);
procedure ShowDebugLines(Lines: String; Caption: String = 'Debug');
function  CommandLineOption(const OptionName: String): Boolean;
function  ExecuteSaveDialogSafely(Dlg: TSaveDialog): Boolean;
function  CreateFileStream(const FileName: String; Mode: Word): TFileStream;

implementation

uses Controls,ComCtrls,StdCtrls;

// adjust these when recompiling with different desktop settings
// DeltaH = Form.Height - Form.ClientHeight
// DeltaW = Form.Width  - Form.ClientWidth   in design time mode.

const
  DESIGN_DELTAH = 27;
  DESIGN_DELTAW =  8;

type
  TAnchorInfo = record
    Control: TControl;
    Anchors: TAnchors;
    OrigXY:  TPoint;
    OrigWH:  TPoint;
  end;
  PAnchorInfo = ^TAnchorInfo;

procedure AdjustFormSize(Form: TForm);
var w,h: Integer;
begin
  w := Form.Width  - DESIGN_DELTAW;
  h := Form.Height - DESIGN_DELTAH;
  if Form.ClientWidth  <> w then Form.ClientWidth  := w;
  if Form.ClientHeight <> h then Form.ClientHeight := h;
end;

procedure ShowDebugLines(Lines,Caption: String);
var
  form:   TForm;
  riched: TRichEdit;
begin
  form := nil;
  try
    form := TForm.Create(Application);
    form.Position := poScreenCenter;
    form.Caption  := Caption;
    form.Width    := 600;
    form.Height   := 400;
    riched := TRichEdit.Create(form);
    riched.Parent := form;
    riched.Font.Name := 'Courier New';
    riched.Font.Size := 8;
    riched.Align := alClient;
    riched.ScrollBars := ssBoth;
    riched.WordWrap := False;
    riched.Text := Lines;
    form.ShowModal;
  finally
    form.Free;
  end;
end;

function  CommandLineOption(const OptionName: String): Boolean;
var i: Integer;
begin
  for i := 1 to ParamCount do
    if CompareText(ParamStr(i), OptionName) = 0 then begin
      Result := True; exit;
    end;
  Result := False;
end;

function  ExecuteSaveDialogSafely(Dlg: TSaveDialog): Boolean;
var t,t0: Int64;
begin
  // dirty workaround if the filename is invalid and save dialog doesn't show
  t0 := GetTickCount; Result := Dlg.Execute;
  t := GetTickCount; if t < t0 then t := t + $100000000;
  if (not Result) and ((t-t0) <= 300) then begin
    Dlg.FileName := '';
    Result := Dlg.Execute;
  end;
end;

function  CreateFileStream(const FileName: String; Mode: Word): TFileStream;
// create a file stream
// if it fails for Creating or Writing: retry a couple of times before giving up
// (may be locked by an AV scanner or by Windows Explorer)
var
  iTry: Integer;
begin
  iTry := 0;
  while True do begin
    inc(iTry);
    try
      Result := TFileStream.Create(FileName, Mode);
      exit;
    except
      if iTry > 3 then
        raise
      else
        Sleep(1000);
    end;
  end;
end;

{ TSafeSaveMemIniFile }

destructor TSafeSaveMemIniFile.Destroy;
begin
  if FQuickWrite then QuickWriteEnd; // frees list
  if FQuickRead  then QuickReadEnd;  // frees list
  inherited;
end;

procedure TSafeSaveMemIniFile.UpdateFile;
var
  OldName,TmpName: String;
begin
  // save to a .tmp file first
  OldName := FileName; TmpName := Filename + '.tmp';
  Rename(TmpName, False); // add .tmp to filename
  try
    inherited UpdateFile; // save data
    DeleteFile(OldName);
    if not RenameFile(TmpName, OldName) then raise Exception.Create('Rename failed for saving ini file');
  finally
    Rename(OldName, False); // rename back to old filename
  end;
end;

procedure TSafeSaveMemIniFile.QuickWriteBegin(const NewSection: String);
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKWRITE} exit; {$ENDIF}
  Assert(not FQuickWrite);
  FQuickWriteStrings := TStringList.Create;
  GetStrings(FQuickWriteStrings);
  FQuickWrite := True;
  FQuickWriteSection := '';
  if NewSection <> '' then QuickWriteNewSection(NewSection);
end;

procedure TSafeSaveMemIniFile.QuickWriteEnd;
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKWRITE} exit; {$ENDIF}
  Assert(FQuickWrite);
  SetStrings(FQuickWriteStrings);
  FQuickWriteStrings.Free;
  FQuickWrite := False;
end;

procedure TSafeSaveMemIniFile.QuickWriteNewSection(const NewSection: String);
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKWRITE} exit; {$ENDIF}
  Assert(FQuickWrite);
  FQuickWriteStrings.Add('[' + NewSection + ']');
  FQuickWriteSection := NewSection;
end;

procedure TSafeSaveMemIniFile.WriteString(const Section, Ident, Value: String);
begin
  if not FQuickWrite then
    inherited
  else begin
    Assert(Section = FQuickWriteSection);
    FQuickWriteStrings.Add(Ident + '=' + Value);
  end;
end;

procedure TSafeSaveMemIniFile.QuickReadBegin(const NewSection: String);
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKREAD} exit; {$ENDIF}
  Assert(not FQuickRead);
  FQuickReadStrings := TStringList.Create;
  FQuickRead := True;
  FQuickReadSection := '';
  FQuickReadIndex   := 0;
  if NewSection <> '' then QuickReadNewSection(NewSection);
end;

procedure TSafeSaveMemIniFile.QuickReadEnd;
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKREAD} exit; {$ENDIF}
  Assert(FQuickRead);
  FQuickReadStrings.Free;
  FQuickRead := False;
end;

procedure TSafeSaveMemIniFile.QuickReadNewSection(const NewSection: String);
begin
  {$IFDEF DEBUG_SSMIF_DISABLE_QUICKREAD} exit; {$ENDIF}
  Assert(FQuickRead);
  ReadSectionValues(NewSection, FQuickReadStrings);
  FQuickReadSection := NewSection;
  FQuickReadIndex   := 0;
end;

function TSafeSaveMemIniFile.ReadString(const Section, Ident, Default: String): String;
var
  ok: Boolean;
begin
  if not FQuickRead then
    Result := inherited ReadString(Section,Ident,Default)
  else begin
    Assert(Section = FQuickReadSection);
    ok := (FQuickReadIndex >= 0) and (FQuickReadIndex < FQuickReadStrings.Count);
    if ok then ok := AnsiCompareText(FQuickReadStrings.Names[FQuickReadIndex], Ident) = 0;
    if ok then begin
      Result := Copy(FQuickReadStrings.Strings[FQuickReadIndex], Length(Ident)+2, MaxInt);
      inc(FQuickReadIndex);
    end else
      Result := inherited ReadString(Section,Ident,Default)
  end;
end;

{ TCachedFile }

constructor TCachedFile.Create(ACacheList: TCachedFileList; AFilename: String; InitFromFile: Boolean);
begin
  inherited Create;
  FCacheList := ACacheList;
  FFilename  := AFilename;
  FMemStream := TMemoryStream.Create;
  FDirty     := False;
  FLastAccess:= Now;
  if InitFromFile then FMemStream.LoadFromFile(FFilename);
  if Assigned(FCacheList) then FCacheList.Add(Self);
end;

destructor TCachedFile.Destroy;
begin
  if Assigned(FCacheList) then FCacheList.Remove(Self);
  FMemStream.Free;
  inherited;
end;

procedure TCachedFile.SetContents(PData: Pointer; Count: Integer);
begin
  FLastAccess:= Now;

  // check if content is identical
  if Count = FMemStream.Size then begin
    if CompareMem(PData, FMemStream.Memory, Count) then exit;
  end;

  FMemStream.SetSize(Count);
  FMemStream.Seek(0, soFromBeginning);
  FMemStream.WriteBuffer(PData^, Count);
  FDirty := True;
end;

procedure TCachedFile.GetContents(PData: Pointer; Count: Integer);
begin
  FLastAccess:= Now;

  if Count <> FMemStream.Size then raise ECachedFile.Create('Size mismatch');
  FMemStream.Seek(0, soFromBeginning);
  FMemStream.ReadBuffer(PData^, Count);
end;

function  TCachedFile.FlushToFile: Boolean;
var Str: TFileStream;
begin
  Result := False;
  if not FDirty then exit;

  Str := CreateFileStream(FFilename, fmCreate or fmShareExclusive);
  try
    FMemStream.SaveToStream(Str);
    FDirty := False;
    Result := True;
  finally
    Str.Free;
  end;
end;

function TCachedFile.GetSize: Int64;
begin
  Result := FMemStream.Size;
end;

{ TCachedFileList }

constructor TCachedFileList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TCachedFileList.Destroy;
var oldcnt: Integer;
begin
  while FList.Count > 0 do begin
    oldcnt := FList.Count;
    TCachedFile(FList.Items[0]).Free; // that removes it from FList too
    if FList.Count = oldcnt then FList.Delete(0); // just in case something failed
  end;
  FList.Free;
  inherited;
end;

procedure TCachedFileList.Add(ACachedFile: TCachedFile);
begin
  FList.Add(ACachedFile);
end;

procedure TCachedFileList.Remove(ACachedFile: TCachedFile);
var idx: Integer;
begin
  idx := FList.IndexOf(ACachedFile);
  if idx >= 0 then FList.Delete(idx);
end;

function TCachedFileList.OpenFile(AFilename: String; MustExist: Boolean): TCachedFile;
var i: Integer;
begin
  for i := 0 to FList.Count-1 do begin
    Result := FList.Items[i];
    if AnsiCompareText(Result.Filename, AFilename) = 0 then exit;
  end;
  Result := TCachedFile.Create(Self, AFilename, MustExist);
end;

function  TCachedFileList.FlushAll(CloseOldFiles: Boolean = False): Integer;
const OLD_FILE_INTERVAL = 30/(24*60); // 30 minutes
var
  i: Integer;
  CachedFile: TCachedFile;
  dt: TDateTime;
begin
  Result := 0;
  dt := Now;
  i := 0;
  while i < FList.Count do begin
    CachedFile := FList.Items[i];
    if CachedFile.FlushToFile then inc(Result);
    if ((dt - CachedFile.LastAccess) > OLD_FILE_INTERVAL) or (dt < CachedFile.LastAccess) then
      CachedFile.Free
    else
      inc(i);
  end;
end;

procedure TCachedFileList.CloseFile(AFilename: String; FlushIt: Boolean);
var
  i:          Integer;
  CachedFile: TCachedFile;
begin
  for i := 0 to FList.Count-1 do begin
    CachedFile := FList.Items[i];
    if AnsiCompareText(CachedFile.Filename, AFilename) = 0 then begin
      if FlushIt then CachedFile.FlushToFile;
      CachedFile.Free;
      exit;
    end;
  end;
end;

procedure TCachedFileList.CloseAll(FlushIt: Boolean);
var CachedFile: TCachedFile;
begin
  while FList.Count > 0 do begin
    CachedFile := FList.Items[0];
    if FlushIt then CachedFile.FlushToFile;
    CachedFile.Free;
  end;
end;

end.
